import time
from typing import Callable
import torch
import math
import os
from tqdm.auto import tqdm
import torch.nn.functional as F
from ailab.atp_dataset.dataset import AILabDataset
from ailab.atp_finetuner.trainer import AILabTrainer 
from ailab.atp_finetuner.model import AILabModel
from ailab.atp_finetuner.datacollator import AILabDataCollator
from ailab.atp_finetuner.metric import AILabMetric
from ailab.atp_finetuner.preprossor import AILabPreprocessor
from ailab.atp_finetuner.build import TrainerRg
from ailab.atp_finetuner.constant import Task, Model
from ailab.log import logger


@TrainerRg.register((Task.text_to_image, Model.stable_diffusion_xl_base_1))
class SDXLTrainer(AILabTrainer):
    def __init__(self):
        super().__init__()

    def preprocess(self, dataset:AILabDataset, model:AILabModel, preprocessor: AILabPreprocessor, \
                      data_collator:AILabDataCollator, metric:AILabMetric, train_progress:Callable, **kwargs):
        train_args = kwargs['train_args']
        self.train_args = train_args
        self.dataset = dataset
        self.model = model
        self.train_progress = train_progress
        self.start_time = time.time()

    def train(self):
        dataloader_num_workers = 0
        lr_warmup_steps = 0
        max_grad_norm = 1
        train_args = self.train_args
        model = self.model
        dataset = self.dataset

        learning_rate = train_args.get('learning_rate', 1e-5)
        gradient_accumulation_steps = train_args.get('gradient_accumulation_steps', 4)
        num_train_epochs = train_args.get('num_train_epochs', 5)
        train_batch_size = train_args.get('per_device_train_batch_size', 1)
        checkpointing_steps = train_args.get('save_steps', 200)
        save_strategy = train_args.get('save_strategy', "no")
        lr_scheduler = train_args.get('lr_scheduler_type', "cosine")
        output_dir = train_args.get('output_dir', "./my_sd_model")
        resume_from_checkpoint = train_args.get('resume_from_checkpoint', False)
        text_encoder_one = model.get_inside_models('text_encoder_one')
        text_encoder_two = model.get_inside_models('text_encoder_two')
        unet = model.get_inside_models('unet')
        vae = model.get_inside_models('vae')
        noise_scheduler = model.get_inside_models('noise_scheduler')
        train_dataset = dataset.to_hf_dataset()
        accelerator = self.accelerator.accelerator_ins

        weight_dtype = torch.float32
        if accelerator.mixed_precision == "fp16":
            weight_dtype = torch.float16
        elif accelerator.mixed_precision == "bf16":
            weight_dtype = torch.bfloat16

        from peft import set_peft_model_state_dict
        from peft.utils import get_peft_model_state_dict
        from diffusers.utils import convert_state_dict_to_diffusers,convert_unet_state_dict_to_peft
        from diffusers import StableDiffusionXLPipeline

        def unwrap_model(model):
            from diffusers.utils.torch_utils import is_compiled_module
            model = accelerator.unwrap_model(model)
            model = model._orig_mod if is_compiled_module(model) else model
            return model

        # create custom saving & loading hooks so that `accelerator.save_state(...)` serializes in a nice format
        def save_model_hook(models, weights, output_dir):
            if accelerator.is_main_process:
                # there are only two options here. Either are just the unet attn processor layers
                # or there are the unet and text encoder attn layers
                unet_lora_layers_to_save = None
                text_encoder_one_lora_layers_to_save = None
                text_encoder_two_lora_layers_to_save = None

                for model in models:
                    if isinstance(unwrap_model(model), type(unwrap_model(unet))):
                        unet_lora_layers_to_save = convert_state_dict_to_diffusers(get_peft_model_state_dict(model))
                    elif isinstance(unwrap_model(model), type(unwrap_model(text_encoder_one))):
                        text_encoder_one_lora_layers_to_save = convert_state_dict_to_diffusers(
                            get_peft_model_state_dict(model)
                        )
                    elif isinstance(unwrap_model(model), type(unwrap_model(text_encoder_two))):
                        text_encoder_two_lora_layers_to_save = convert_state_dict_to_diffusers(
                            get_peft_model_state_dict(model)
                        )
                    else:
                        raise ValueError(f"unexpected save model: {model.__class__}")

                    # make sure to pop weight so that corresponding model is not saved again
                    if weights:
                        weights.pop()

                StableDiffusionXLPipeline.save_lora_weights(
                    output_dir,
                    unet_lora_layers=unet_lora_layers_to_save,
                    text_encoder_lora_layers=text_encoder_one_lora_layers_to_save,
                    text_encoder_2_lora_layers=text_encoder_two_lora_layers_to_save,
                )

        def load_model_hook(models, input_dir):
            unet_ = None
            text_encoder_one_ = None
            text_encoder_two_ = None

            while len(models) > 0:
                model = models.pop()

                if isinstance(model, type(unwrap_model(unet))):
                    unet_ = model
                elif isinstance(model, type(unwrap_model(text_encoder_one))):
                    text_encoder_one_ = model
                elif isinstance(model, type(unwrap_model(text_encoder_two))):
                    text_encoder_two_ = model
                else:
                    raise ValueError(f"unexpected save model: {model.__class__}")
            
            from diffusers.loaders import LoraLoaderMixin
            lora_state_dict, _ = LoraLoaderMixin.lora_state_dict(input_dir)
            unet_state_dict = {f'{k.replace("unet.", "")}': v for k, v in lora_state_dict.items() if k.startswith("unet.")}
            unet_state_dict = convert_unet_state_dict_to_peft(unet_state_dict)
            incompatible_keys = set_peft_model_state_dict(unet_, unet_state_dict, adapter_name="default")
            if incompatible_keys is not None:
                # check only for unexpected keys
                unexpected_keys = getattr(incompatible_keys, "unexpected_keys", None)
                if unexpected_keys:
                    print(
                        f"Loading adapter weights from state_dict led to unexpected keys not found in the model: "
                        f" {unexpected_keys}. "
                    )

        accelerator.register_save_state_pre_hook(save_model_hook)
        accelerator.register_load_state_pre_hook(load_model_hook)

        from diffusers.loaders import AttnProcsLayers
        from diffusers.optimization import get_scheduler
        optimizer_class = torch.optim.AdamW
        params_to_optimize = list(filter(lambda p: p.requires_grad, unet.parameters()))
        optimizer = optimizer_class(
            params_to_optimize,
            lr=learning_rate,
            betas=(0.9, 0.999),
            weight_decay=0.01,
            eps=1e-08,
        )

        def collate_fn(examples):
            pixel_values = torch.stack([example["pixel_values"] for example in examples])
            pixel_values = pixel_values.to(memory_format=torch.contiguous_format).float()
            original_sizes = [example["original_sizes"] for example in examples]
            crop_top_lefts = [example["crop_top_lefts"] for example in examples]
            input_ids_one = torch.stack([example["input_ids_one"] for example in examples])
            input_ids_two = torch.stack([example["input_ids_two"] for example in examples])
            result = {
                "pixel_values": pixel_values,
                "input_ids_one": input_ids_one,
                "input_ids_two": input_ids_two,
                "original_sizes": original_sizes,
                "crop_top_lefts": crop_top_lefts,
            }

            filenames = [example["filenames"] for example in examples if "filenames" in example]
            if filenames:
                result["filenames"] = filenames
            return result

        # DataLoaders creation:
        train_dataloader = torch.utils.data.DataLoader(
            train_dataset,
            shuffle=True,
            collate_fn=collate_fn,
            batch_size=train_batch_size,
            num_workers=dataloader_num_workers,
        )

        # Scheduler and math around the number of training steps.
        num_update_steps_per_epoch = math.ceil(len(train_dataloader) / gradient_accumulation_steps)
        max_train_steps = num_train_epochs * num_update_steps_per_epoch

        from diffusers.optimization import get_scheduler
        lr_warmup_steps=500
        lr_scheduler = get_scheduler(
            'constant',
            optimizer=optimizer,
            num_warmup_steps=lr_warmup_steps * gradient_accumulation_steps,
            num_training_steps=max_train_steps * gradient_accumulation_steps,
        )
        unet, optimizer, train_dataloader, lr_scheduler = accelerator.prepare(
                unet, optimizer, train_dataloader, lr_scheduler
            )

        # We need to recalculate our total training steps as the size of the training dataloader may have changed.
        num_update_steps_per_epoch = math.ceil(len(train_dataloader) / gradient_accumulation_steps)
        max_train_steps = num_train_epochs * num_update_steps_per_epoch
        # Afterwards we recalculate our number of training epochs
        num_train_epochs = math.ceil(max_train_steps / num_update_steps_per_epoch)

        # We need to initialize the trackers we use, and also store our configuration.
        # The trackers initializes automatically on the main process.
        if accelerator.is_main_process:
            accelerator.init_trackers("text2image-fine-tune")
        

        total_batch_size = train_batch_size * accelerator.num_processes * gradient_accumulation_steps
        if accelerator.is_main_process:
            print(train_dataset)
            logger.info("***** Running training *****")
            logger.info(f"  Num examples = {len(train_dataset)}")
            logger.info(f"  Num Epochs = {num_train_epochs}")
            logger.info(f"  Instantaneous batch size per device = {train_batch_size}")
            logger.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
            logger.info(f"  Gradient Accumulation steps = {gradient_accumulation_steps}")
            logger.info(f"  Total optimization steps = {max_train_steps}")
        global_step = 0
        first_epoch = 0

        if resume_from_checkpoint:
            # Get the most recent checkpoint
            # Get the most recent checkpoint
            dirs = os.listdir(output_dir)
            dirs = [d for d in dirs if d.startswith("checkpoint")]
            dirs = sorted(dirs, key=lambda x: int(x.split("-")[1]))
            path = dirs[-1] if len(dirs) > 0 else None

            if path is None:
                resume_from_checkpoint = None
                initial_global_step = 0
            else:
                logger.info(f"Resuming from checkpoint {resume_from_checkpoint}")
                accelerator.load_state(os.path.join(output_dir, path))
                global_step = int(path.split("-")[1])

                #resume_global_step = global_step * gradient_accumulation_steps
                first_epoch = global_step // num_update_steps_per_epoch
                initial_global_step = global_step
                #resume_step = resume_global_step % (num_update_steps_per_epoch * gradient_accumulation_steps)
        else:
            initial_global_step = 0
        # Only show the progress bar once on each machine.
        progress_bar = tqdm(
            range(0, max_train_steps), 
            initial=initial_global_step,
            desc="Steps",
            disable=not accelerator.is_local_main_process)
        progress_bar.set_description("Steps")

        def tokenize_prompt(tokenizer, prompt):
            text_inputs = tokenizer(
                prompt,
                padding="max_length",
                max_length=tokenizer.model_max_length,
                truncation=True,
                return_tensors="pt",
            )
            text_input_ids = text_inputs.input_ids
            return text_input_ids

        def encode_prompt(text_encoders, tokenizers, prompt, text_input_ids_list=None):
            prompt_embeds_list = []

            for i, text_encoder in enumerate(text_encoders):
                if tokenizers is not None:
                    tokenizer = tokenizers[i]
                    text_input_ids = tokenize_prompt(tokenizer, prompt)
                else:
                    assert text_input_ids_list is not None
                    text_input_ids = text_input_ids_list[i]

                prompt_embeds = text_encoder(
                    text_input_ids.to(text_encoder.device), output_hidden_states=True, return_dict=False
                )

                # We are only ALWAYS interested in the pooled output of the final text encoder
                pooled_prompt_embeds = prompt_embeds[0]
                prompt_embeds = prompt_embeds[-1][-2]
                bs_embed, seq_len, _ = prompt_embeds.shape
                prompt_embeds = prompt_embeds.view(bs_embed, seq_len, -1)
                prompt_embeds_list.append(prompt_embeds)

            prompt_embeds = torch.concat(prompt_embeds_list, dim=-1)
            pooled_prompt_embeds = pooled_prompt_embeds.view(bs_embed, -1)
            return prompt_embeds, pooled_prompt_embeds

        step_start = time.time()
        for epoch in range(first_epoch, num_train_epochs):
            unet.train()
            train_loss = 0.0
            for step, batch in enumerate(train_dataloader):
                with accelerator.accumulate(unet):
                    # Convert images to latent space
                    pixel_values = batch["pixel_values"]

                    model_input = vae.encode(pixel_values).latent_dist.sample()
                    model_input = model_input * vae.config.scaling_factor
                    model_input = model_input.to(weight_dtype)

                    # Sample noise that we'll add to the latents
                    noise = torch.randn_like(model_input)
                    bsz = model_input.shape[0]
                    # Sample a random timestep for each image
                    timesteps = torch.randint(
                        0, noise_scheduler.config.num_train_timesteps, (bsz,), device=model_input.device
                    )
                    timesteps = timesteps.long()

                    # Add noise to the model input according to the noise magnitude at each timestep
                    # (this is the forward diffusion process)
                    noisy_model_input = noise_scheduler.add_noise(model_input, noise, timesteps)

                    # time ids
                    resolution = 1024
                    def compute_time_ids(original_size, crops_coords_top_left):
                        # Adapted from pipeline.StableDiffusionXLPipeline._get_add_time_ids
                        target_size = (resolution, resolution)
                        add_time_ids = list(original_size + crops_coords_top_left + target_size)
                        add_time_ids = torch.tensor([add_time_ids])
                        add_time_ids = add_time_ids.to(accelerator.device, dtype=weight_dtype)
                        return add_time_ids

                    add_time_ids = torch.cat(
                        [compute_time_ids(s, c) for s, c in zip(batch["original_sizes"], batch["crop_top_lefts"])]
                    )

                    # Predict the noise residual
                    unet_added_conditions = {"time_ids": add_time_ids}
                    prompt_embeds, pooled_prompt_embeds = encode_prompt(
                        text_encoders=[text_encoder_one, text_encoder_two],
                        tokenizers=None,
                        prompt=None,
                        text_input_ids_list=[batch["input_ids_one"], batch["input_ids_two"]],
                    )
                    unet_added_conditions.update({"text_embeds": pooled_prompt_embeds})
                    model_pred = unet(
                        noisy_model_input,
                        timesteps,
                        prompt_embeds,
                        added_cond_kwargs=unet_added_conditions,
                        return_dict=False,
                    )[0]

                    if noise_scheduler.config.prediction_type == "epsilon":
                        target = noise
                    elif noise_scheduler.config.prediction_type == "v_prediction":
                        target = noise_scheduler.get_velocity(model_input, noise, timesteps)
                    else:
                        raise ValueError(f"Unknown prediction type {noise_scheduler.config.prediction_type}")

                    import torch.nn.functional as F
                    loss = F.mse_loss(model_pred.float(), target.float(), reduction="mean")
                    # Gather the losses across all processes for logging (if we use distributed training).
                    avg_loss = accelerator.gather(loss.repeat(train_batch_size)).mean()
                    train_loss += avg_loss.item() / gradient_accumulation_steps

                    max_grad_norm = 1.0
                    # Backpropagate
                    accelerator.backward(loss)
                    if accelerator.sync_gradients:
                        accelerator.clip_grad_norm_(params_to_optimize, max_grad_norm)
                    optimizer.step()
                    lr_scheduler.step()
                    optimizer.zero_grad()

                # Checks if the accelerator has performed an optimization step behind the scenes
                if accelerator.sync_gradients:
                    progress_bar.update(1)
                    global_step += 1
                    accelerator.log({"train_loss": train_loss}, step=global_step)
                    train_loss = 0.0

                logs = {"step_loss": loss.detach().item(), "lr": lr_scheduler.get_last_lr()[0]}
                progress_bar.set_postfix(**logs)

                if global_step >= max_train_steps:
                    break

        # Save the lora layers
        accelerator.wait_for_everyone()
        if accelerator.is_main_process:
            unet = unwrap_model(unet)
            from diffusers import StableDiffusionXLPipeline
            unet_lora_state_dict = convert_state_dict_to_diffusers(get_peft_model_state_dict(unet))

            text_encoder_lora_layers = None
            text_encoder_2_lora_layers = None

            StableDiffusionXLPipeline.save_lora_weights(
                save_directory=output_dir,
                unet_lora_layers=unet_lora_state_dict,
                text_encoder_lora_layers=text_encoder_lora_layers,
                text_encoder_2_lora_layers=text_encoder_2_lora_layers,
            )
            end_to_zip = self.train_args.get('end_to_zip', False)
            if end_to_zip:
                from ailab.utils.other import create_zip_and_delete_folder
                zip_file_path = output_dir+"/adapter.zip"
                create_zip_and_delete_folder(output_dir,zip_file_path)

        accelerator.end_training()

    def postprocess(self):
        pass



